@@ -2,7 +2,7 @@
 #
 # autoblockchainify — Turn a directory into a GIT Blockchain
 #
-# Copyright (C) 2019,2020 Marcel Waldvogel
+# Copyright (C) 2019-2021 Marcel Waldvogel
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License as published
@@ -21,7 +21,7 @@
 # Set up the daemon
 
 
-import logging as _logging
+import signale
 import subprocess
 from pathlib import Path
 
@@ -29,7 +29,13 @@ import autoblockchainify.commit
 import autoblockchainify.config
 import autoblockchainify.version
 
-logging = _logging.getLogger('daemon')
+
+logging = signale.Signale({"scope": "daemon"})
+try:
+    # Aligned output, if supported by this version of Signale
+    signale.set_align(8, 8)
+except AttributeError:
+    pass
 
 
 def finish_setup(arg):
@@ -50,7 +56,7 @@ def finish_setup(arg):
 .ssh/
 ''')
     if arg.identity is not None:
-        logging.info("Initializing new repo with user info")
+        logging.start("Initializing new repo with user info")
         (name, mail) = arg.identity.split(' <')
         subprocess.run(['git', 'config', 'user.name', name],
                        cwd=repo, check=True)
@@ -63,7 +69,6 @@ def run():
     finish_setup(autoblockchainify.config.arg)
     # Try to resume waiting for a PGP Timestamping Server reply, if any
     if autoblockchainify.config.arg.stamper_own_address:
-        repo = autoblockchainify.config.arg.repository
-        logging.info("possibly resuming cross-timestamping by mail")
+        logging.pending("possibly resuming cross-timestamping by mail")
         autoblockchainify.mail.async_email_timestamp(resume=True)
     autoblockchainify.commit.loop()
